"""Logger dependency for FastAPI.

Provides a `structlog` logger as a FastAPI dependency.  The logger will
incorporate information from the request in its bound context.
"""

import uuid
from typing import Optional

import structlog
from fastapi import Request
from structlog.stdlib import BoundLogger

from safir import logging

__all__ = ["LoggerDependency", "logger_dependency"]


class LoggerDependency:
    """Provides a structlog logger configured with request information.

    The following additional information will be included:

    * A UUID for the request
    * The method and URL of the request
    * The IP address of the client
    * The ``User-Agent`` header of the request, if any.

    The last three pieces of information will be added using naming consistent
    with the expectations of Google Log Explorer so that the request
    information will be liftedn into the appropriate JSON fields for complex
    log queries.
    """

    def __init__(self) -> None:
        self.logger: Optional[BoundLogger] = None

    async def __call__(self, request: Request) -> BoundLogger:
        """Return a logger bound with request information.

        Returns
        -------
        logger : `structlog.stdlib.BoundLogger`
            The bound logger.
        """
        if not self.logger:
            self.logger = structlog.get_logger(logging.logger_name)
        assert self.logger

        # Construct the httpRequest logging data (compatible with the format
        # expected by Google Log Explorer).
        request_data = {
            "requestMethod": request.method,
            "requestUrl": str(request.url),
            "remoteIp": request.client.host,
        }
        user_agent = request.headers.get("User-Agent")
        if user_agent:
            request_data["userAgent"] = user_agent

        logger = self.logger.new(
            httpRequest=request_data,
            request_id=str(uuid.uuid4()),
        )
        return logger


logger_dependency = LoggerDependency()
"""The dependency that will return the logger for the current request."""



"""Fixtures for tests of the testing support infrastructure.

Defines fixtures similar to what users of the Safir library should define so
that they can be exercised in tests of the test support infrastructure.
"""

from __future__ import annotations

from typing import Iterator

import pytest

from safir.testing.kubernetes import MockKubernetesApi, patch_kubernetes


@pytest.fixture
def mock_kubernetes() -> Iterator[MockKubernetesApi]:
    yield from patch_kubernetes()



"""Tests for the Kubernetes support infrastructure.

These are just basic sanity checks that the mocking is working correctly and
the basic calls work.
"""

from __future__ import annotations

from typing import Any, Dict
from unittest.mock import ANY

import pytest
from kubernetes_asyncio import config
from kubernetes_asyncio.client import V1ObjectMeta, V1Secret

from safir.kubernetes import initialize_kubernetes
from safir.testing.kubernetes import MockKubernetesApi


@pytest.mark.asyncio
async def test_initialize(mock_kubernetes: MockKubernetesApi) -> None:
    await initialize_kubernetes()
    assert config.load_incluster_config.call_count == 1


@pytest.mark.asyncio
async def test_mock(mock_kubernetes: MockKubernetesApi) -> None:
    custom: Dict[str, Any] = {
        "apiVersion": "gafaelfawr.lsst.io/v1alpha1",
        "kind": "GafaelfawrServiceToken",
        "metadata": {
            "name": "gafaelfawr-secret",
            "namespace": "mobu",
            "generation": 1,
        },
        "spec": {
            "service": "mobu",
            "scopes": ["admin:token"],
        },
    }
    secret = V1Secret(
        api_version="v1",
        kind="Secret",
        data={"token": "bogus"},
        metadata=V1ObjectMeta(name="gafaelfawr-secret", namespace="mobu"),
        type="Opaque",
    )
    await mock_kubernetes.create_namespaced_custom_object(
        "gafaelfawr.lsst.io",
        "v1alpha1",
        custom["metadata"]["namespace"],
        "gafaelfawrservicetokens",
        custom,
    )
    await mock_kubernetes.create_namespaced_secret(
        secret.metadata.namespace, secret
    )

    assert await mock_kubernetes.list_cluster_custom_object(
        "gafaelfawr.lsst.io", "v1alpha1", "gafaelfawrservicetokens"
    ) == {
        "items": [{**custom, "metadata": {**custom["metadata"], "uid": ANY}}]
    }
    assert mock_kubernetes.get_all_objects_for_test("Secret") == [secret]

    def error(method: str, *args: Any) -> None:
        assert method == "replace_namespaced_custom_object"
        raise ValueError("some exception")

    mock_kubernetes.error_callback = error
    with pytest.raises(ValueError) as excinfo:
        await mock_kubernetes.replace_namespaced_custom_object(
            "gafaelfawr.lsst.io",
            "v1alpha1",
            custom["metadata"]["namespace"],
            "gafaelfawrservicetokens",
            "gafaelfawr-secret",
            custom,
        )
    assert str(excinfo.value) == "some exception"



"""Standard models for FastAPI applications.

Examples
--------
To reference the `ErrorModel` model when returning an error message, use code
similar to this:

.. code-block:: python

   @router.get(
       "/route/{foo}",
       ...,
       responses={404: {"description": "Not found", "model": ErrorModel}},
   )
   async def route(foo: str) -> None:
       ...
       raise HTTPException(
           status_code=status.HTTP_404_NOT_FOUND,
           detail=[
               {"loc": ["path", "foo"], "msg": msg, "type": "invalid_foo"},
           ],
       )

Notes
-----
FastAPI does not appear to export its error response model in a usable form,
so define a copy of it so that we can reference it in API definitions to
generate good documentation.
"""

# Examples and notes kept in the module docstring because they're not
# appropriate for the API documentation generated for a service.

from typing import List, Optional

from pydantic import BaseModel, Field

__all__ = ["ErrorDetail", "ErrorModel"]


class ErrorDetail(BaseModel):
    """The detail of the error message."""

    loc: Optional[List[str]] = Field(
        None, title="Location", example=["area", "field"]
    )

    msg: str = Field(..., title="Message", example="Some error messge")

    type: str = Field(..., title="Error type", example="some_code")


class ErrorModel(BaseModel):
    """A structured API error message."""

    detail: List[ErrorDetail] = Field(..., title="Detail")
